/*
    Copyright 2013 Mega-Mario

    This file is part of lolSnes.

    lolSnes is free software: you can redistribute it and/or modify it under
    the terms of the GNU General Public License as published by the Free
    Software Foundation, either version 3 of the License, or (at your option)
    any later version.

    lolSnes is distributed in the hope that it will be useful, but WITHOUT ANY 
    WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
    FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along 
    with lolSnes. If not, see http://www.gnu.org/licenses/.
*/

#ifndef PPU_PRIO_H
#define PPU_PRIO_H

/*
	PPU priority tables
	
	On the SNES, the main/sub screen feature can alter how layers are ordered.
	The PPU basically renders the main screen layers with proper priorities and
	isolates the topmost main screen pixel, then does the same for the sub screen,
	and mixes the two pixels according to the color math registers.
	
	For example, SMW has BG1, BG3 and OBJ on main screen, and BG2 on sub screen.
	CGADSUB is configured so that sub screen pixels are only mixed with main screen
	pixels if they're from the backdrop.
	
	Normally, the layer order would be BG1 > BG2 > BG3. But with main/sub screen 
	stuff in mind, BG2 only appears where other layers don't. The actual order is
	BG1 > BG3 > BG2.
	
	To reproduce this on the DS, since we don't have main/sub screens there, we
	need to assign priority to main screen layers and sub screen layers separately.
	Given we only have 4 priority slots to work with, handling that automatically
	would be too complex.
	
	As a workaround, these tables define priority values for all the possible cases.
	
	(note: mode 8 is mode 1 with high BG3 prio)
	
	(note: lines marked with !! mean that more than 4 priority slots would be needed.
	Typically tile prio for farthest BGs is sacrificed in those cases)
	
	16 bytes per entry
	* index of the DS layer each SNES layer is mapped to
	* layer priority values for low-prio tiles
	* layer priority values for high-prio tiles
	* OBJ priority values
*/

u8 PPU_PrioTable[9 * 32 * 16] = 
{
	// Mode 0
	
	2, 3, 0, 1,   1, 1, 3, 3,   0, 0, 2, 2,   3, 2, 1, 0,
	2, 3, 0, 1,   3, 1, 3, 3,   3, 0, 2, 2,   3, 2, 1, 0,
	2, 3, 0, 1,   1, 3, 3, 3,   0, 3, 2, 2,   3, 2, 1, 0,
	2, 3, 0, 1,   3, 3, 1, 1,   2, 2, 0, 0,   1, 0, 0, 0,
	1, 2, 3, 0,   1, 1, 3, 3,   0, 0, 3, 2,   3, 2, 1, 0,
	1, 2, 3, 0,   3, 1, 3, 3,   3, 0, 3, 2,   3, 2, 1, 0,
	1, 2, 3, 0,   1, 3, 3, 3,   0, 3, 3, 2,   3, 2, 1, 0,
	1, 2, 3, 0,   2, 2, 3, 1,   1, 1, 3, 0,   1, 0, 0, 0,
	1, 2, 0, 3,   1, 1, 3, 3,   0, 0, 2, 3,   3, 2, 1, 0,
	1, 2, 0, 3,   3, 1, 3, 3,   3, 0, 2, 3,   3, 2, 1, 0,
	1, 2, 0, 3,   1, 3, 3, 3,   0, 3, 2, 3,   3, 2, 1, 0,
	1, 2, 0, 3,   2, 2, 1, 3,   1, 1, 0, 3,   1, 0, 0, 0,
	0, 1, 2, 3,   1, 1, 3, 3,   0, 0, 2, 2,   2, 2, 1, 0,
	0, 1, 2, 3,   2, 1, 3, 3,   2, 0, 2, 2,   2, 2, 1, 0,
	0, 1, 2, 3,   1, 2, 3, 3,   0, 2, 2, 2,   2, 2, 1, 0,
	0, 1, 2, 3,   1, 1, 3, 3,   0, 0, 2, 2,   0, 0, 0, 0,
	
	0, 1, 2, 3,   1, 1, 2, 2,   0, 0, 1, 1,   3, 3, 3, 3,
	0, 1, 2, 3,   2, 0, 1, 1,   2, 0, 0, 0,   3, 3, 2, 2, // !!
	0, 1, 2, 3,   0, 2, 1, 1,   0, 2, 0, 0,   3, 3, 2, 2, // !!
	0, 1, 2, 3,   2, 2, 1, 1,   2, 2, 0, 0,   3, 3, 2, 2, // !!
	0, 1, 2, 3,   1, 1, 3, 1,   0, 0, 2, 1,   3, 2, 2, 2,
	0, 1, 2, 3,   2, 0, 3, 0,   1, 0, 3, 0,   3, 3, 2, 1, // !!
	0, 1, 2, 3,   0, 2, 3, 0,   0, 1, 3, 0,   3, 3, 2, 1, // !!
	0, 1, 2, 3,   2, 2, 3, 0,   1, 1, 3, 0,   3, 3, 2, 1, // !!
	0, 1, 2, 3,   1, 1, 1, 3,   0, 0, 1, 2,   3, 2, 2, 2,
	0, 1, 2, 3,   2, 0, 0, 3,   1, 0, 0, 3,   3, 3, 2, 1, // !!
	0, 1, 2, 3,   0, 2, 0, 3,   0, 1, 0, 3,   3, 3, 2, 1, // !!
	0, 1, 2, 3,   2, 2, 0, 3,   1, 1, 0, 3,   3, 3, 2, 1, // !!
	0, 1, 2, 3,   1, 1, 3, 3,   0, 0, 2, 2,   3, 2, 2, 2, 
	0, 1, 2, 3,   2, 0, 3, 3,   1, 0, 3, 3,   3, 3, 2, 1, // !!
	0, 1, 2, 3,   0, 2, 3, 3,   0, 1, 3, 3,   3, 3, 2, 1, // !!
	0, 1, 2, 3,   1, 1, 3, 3,   0, 0, 2, 2,   3, 2, 1, 0,
	
	// Mode 1
	
	1, 2, 0, 3,   1, 1, 3, 3,   0, 0, 2, 3,   3, 2, 1, 0,
	1, 2, 0, 3,   3, 1, 3, 3,   3, 0, 2, 3,   3, 2, 1, 0,
	1, 2, 0, 3,   1, 3, 3, 3,   0, 3, 2, 3,   3, 2, 1, 0,
	1, 2, 0, 3,   3, 3, 1, 3,   2, 2, 0, 3,   1, 0, 0, 0,
	1, 2, 0, 3,   1, 1, 3, 3,   0, 0, 3, 3,   3, 2, 1, 0,
	1, 2, 0, 3,   3, 1, 3, 3,   3, 0, 3, 3,   3, 2, 1, 0,
	1, 2, 0, 3,   1, 3, 3, 3,   0, 3, 3, 3,   3, 2, 1, 0,
	1, 2, 0, 3,   2, 2, 3, 3,   1, 1, 3, 3,   1, 0, 0, 0,
	1, 2, 0, 3,   1, 1, 3, 3,   0, 0, 2, 3,   3, 2, 1, 0,
	1, 2, 0, 3,   3, 1, 3, 3,   3, 0, 2, 3,   3, 2, 1, 0,
	1, 2, 0, 3,   1, 3, 3, 3,   0, 3, 2, 3,   3, 2, 1, 0,
	1, 2, 0, 3,   3, 3, 1, 3,   2, 2, 0, 3,   1, 0, 0, 0,
	1, 2, 0, 3,   1, 1, 3, 3,   0, 0, 3, 3,   3, 2, 1, 0,
	1, 2, 0, 3,   3, 1, 3, 3,   3, 0, 3, 3,   3, 2, 1, 0,
	1, 2, 0, 3,   1, 3, 3, 3,   0, 3, 3, 3,   3, 2, 1, 0,
	1, 2, 0, 3,   2, 2, 3, 3,   1, 1, 3, 3,   1, 0, 0, 0,
	
	0, 1, 2, 3,   1, 1, 2, 3,   0, 0, 2, 3,   3, 3, 3, 3, 
	0, 1, 2, 3,   2, 0, 0, 3,   1, 0, 0, 3,   3, 3, 2, 1,
	0, 1, 2, 3,   0, 2, 0, 3,   0, 1, 0, 3,   3, 3, 2, 1,
	0, 1, 2, 3,   2, 2, 0, 3,   1, 1, 0, 3,   3, 3, 2, 1,
	0, 1, 2, 3,   1, 1, 3, 3,   0, 0, 2, 3,   3, 2, 2, 2,
	0, 1, 2, 3,   2, 0, 3, 3,   1, 0, 3, 3,   3, 3, 2, 1, // !!
	0, 1, 2, 3,   0, 2, 3, 3,   0, 1, 3, 3,   3, 3, 2, 1, // !!
	0, 1, 2, 3,   1, 1, 3, 3,   0, 0, 2, 3,   3, 2, 1, 0,
	0, 1, 2, 3,   1, 1, 2, 3,   0, 0, 2, 3,   3, 3, 3, 3, 
	0, 1, 2, 3,   2, 0, 0, 3,   1, 0, 0, 3,   3, 3, 2, 1,
	0, 1, 2, 3,   0, 2, 0, 3,   0, 1, 0, 3,   3, 3, 2, 1,
	0, 1, 2, 3,   2, 2, 0, 3,   1, 1, 0, 3,   3, 3, 2, 1,
	0, 1, 2, 3,   1, 1, 3, 3,   0, 0, 2, 3,   3, 2, 2, 2,
	0, 1, 2, 3,   2, 0, 3, 3,   1, 0, 3, 3,   3, 3, 2, 1, // !!
	0, 1, 2, 3,   0, 2, 3, 3,   0, 1, 3, 3,   3, 3, 2, 1, // !!
	0, 1, 2, 3,   1, 1, 3, 3,   0, 0, 2, 3,   3, 2, 1, 0,
	
	// Mode 2
	
	1, 2, 0, 3,   2, 3, 3, 3,   0, 1, 3, 3,   3, 2, 1, 0,
	1, 2, 0, 3,   2, 1, 3, 3,   2, 0, 3, 3,   1, 1, 0, 0,
	1, 2, 0, 3,   1, 2, 3, 3,   0, 2, 3, 3,   2, 1, 1, 0,
	1, 2, 0, 3,   1, 1, 3, 3,   0, 0, 3, 3,   0, 0, 0, 0,
	1, 2, 0, 3,   2, 3, 3, 3,   0, 1, 3, 3,   3, 2, 1, 0,
	1, 2, 0, 3,   2, 1, 3, 3,   2, 0, 3, 3,   1, 1, 0, 0,
	1, 2, 0, 3,   1, 2, 3, 3,   0, 2, 3, 3,   2, 1, 1, 0,
	1, 2, 0, 3,   1, 1, 3, 3,   0, 0, 3, 3,   0, 0, 0, 0,
	1, 2, 0, 3,   2, 3, 3, 3,   0, 1, 3, 3,   3, 2, 1, 0,
	1, 2, 0, 3,   2, 1, 3, 3,   2, 0, 3, 3,   1, 1, 0, 0,
	1, 2, 0, 3,   1, 2, 3, 3,   0, 2, 3, 3,   2, 1, 1, 0,
	1, 2, 0, 3,   1, 1, 3, 3,   0, 0, 3, 3,   0, 0, 0, 0,
	1, 2, 0, 3,   2, 3, 3, 3,   0, 1, 3, 3,   3, 2, 1, 0,
	1, 2, 0, 3,   2, 1, 3, 3,   2, 0, 3, 3,   1, 1, 0, 0,
	1, 2, 0, 3,   1, 2, 3, 3,   0, 2, 3, 3,   2, 1, 1, 0,
	1, 2, 0, 3,   1, 1, 3, 3,   0, 0, 3, 3,   0, 0, 0, 0,
	
	1, 2, 0, 3,   2, 2, 3, 3,   1, 1, 3, 3,   3, 3, 3, 3,
	1, 2, 0, 3,   2, 0, 3, 3,   1, 0, 3, 3,   3, 2, 2, 1,
	1, 2, 0, 3,   1, 3, 3, 3,   1, 2, 3, 3,   3, 3, 2, 2,
	1, 2, 0, 3,   2, 3, 3, 3,   0, 1, 3, 3,   3, 2, 1, 0,
	1, 2, 0, 3,   2, 2, 3, 3,   1, 1, 3, 3,   3, 3, 3, 3,
	1, 2, 0, 3,   2, 0, 3, 3,   1, 0, 3, 3,   3, 2, 2, 1,
	1, 2, 0, 3,   1, 3, 3, 3,   1, 2, 3, 3,   3, 3, 2, 2,
	1, 2, 0, 3,   2, 3, 3, 3,   0, 1, 3, 3,   3, 2, 1, 0,
	1, 2, 0, 3,   2, 2, 3, 3,   1, 1, 3, 3,   3, 3, 3, 3,
	1, 2, 0, 3,   2, 0, 3, 3,   1, 0, 3, 3,   3, 2, 2, 1,
	1, 2, 0, 3,   1, 3, 3, 3,   1, 2, 3, 3,   3, 3, 2, 2,
	1, 2, 0, 3,   2, 3, 3, 3,   0, 1, 3, 3,   3, 2, 1, 0,
	1, 2, 0, 3,   2, 2, 3, 3,   1, 1, 3, 3,   3, 3, 3, 3,
	1, 2, 0, 3,   2, 0, 3, 3,   1, 0, 3, 3,   3, 2, 2, 1,
	1, 2, 0, 3,   1, 3, 3, 3,   1, 2, 3, 3,   3, 3, 2, 2,
	1, 2, 0, 3,   2, 3, 3, 3,   0, 1, 3, 3,   3, 2, 1, 0,
	
	// Mode 3
	
	1, 2, 0, 3,   2, 3, 3, 3,   0, 1, 3, 3,   3, 2, 1, 0,
	1, 2, 0, 3,   2, 1, 3, 3,   2, 0, 3, 3,   1, 1, 0, 0,
	1, 2, 0, 3,   1, 2, 3, 3,   0, 2, 3, 3,   2, 1, 1, 0,
	1, 2, 0, 3,   1, 1, 3, 3,   0, 0, 3, 3,   0, 0, 0, 0,
	1, 2, 0, 3,   2, 3, 3, 3,   0, 1, 3, 3,   3, 2, 1, 0,
	1, 2, 0, 3,   2, 1, 3, 3,   2, 0, 3, 3,   1, 1, 0, 0,
	1, 2, 0, 3,   1, 2, 3, 3,   0, 2, 3, 3,   2, 1, 1, 0,
	1, 2, 0, 3,   1, 1, 3, 3,   0, 0, 3, 3,   0, 0, 0, 0,
	1, 2, 0, 3,   2, 3, 3, 3,   0, 1, 3, 3,   3, 2, 1, 0,
	1, 2, 0, 3,   2, 1, 3, 3,   2, 0, 3, 3,   1, 1, 0, 0,
	1, 2, 0, 3,   1, 2, 3, 3,   0, 2, 3, 3,   2, 1, 1, 0,
	1, 2, 0, 3,   1, 1, 3, 3,   0, 0, 3, 3,   0, 0, 0, 0,
	1, 2, 0, 3,   2, 3, 3, 3,   0, 1, 3, 3,   3, 2, 1, 0,
	1, 2, 0, 3,   2, 1, 3, 3,   2, 0, 3, 3,   1, 1, 0, 0,
	1, 2, 0, 3,   1, 2, 3, 3,   0, 2, 3, 3,   2, 1, 1, 0,
	1, 2, 0, 3,   1, 1, 3, 3,   0, 0, 3, 3,   0, 0, 0, 0,
	
	1, 2, 0, 3,   2, 2, 3, 3,   1, 1, 3, 3,   3, 3, 3, 3,
	1, 2, 0, 3,   2, 0, 3, 3,   1, 0, 3, 3,   3, 2, 2, 1,
	1, 2, 0, 3,   1, 3, 3, 3,   1, 2, 3, 3,   3, 3, 2, 2,
	1, 2, 0, 3,   2, 3, 3, 3,   0, 1, 3, 3,   3, 2, 1, 0,
	1, 2, 0, 3,   2, 2, 3, 3,   1, 1, 3, 3,   3, 3, 3, 3,
	1, 2, 0, 3,   2, 0, 3, 3,   1, 0, 3, 3,   3, 2, 2, 1,
	1, 2, 0, 3,   1, 3, 3, 3,   1, 2, 3, 3,   3, 3, 2, 2,
	1, 2, 0, 3,   2, 3, 3, 3,   0, 1, 3, 3,   3, 2, 1, 0,
	1, 2, 0, 3,   2, 2, 3, 3,   1, 1, 3, 3,   3, 3, 3, 3,
	1, 2, 0, 3,   2, 0, 3, 3,   1, 0, 3, 3,   3, 2, 2, 1,
	1, 2, 0, 3,   1, 3, 3, 3,   1, 2, 3, 3,   3, 3, 2, 2,
	1, 2, 0, 3,   2, 3, 3, 3,   0, 1, 3, 3,   3, 2, 1, 0,
	1, 2, 0, 3,   2, 2, 3, 3,   1, 1, 3, 3,   3, 3, 3, 3,
	1, 2, 0, 3,   2, 0, 3, 3,   1, 0, 3, 3,   3, 2, 2, 1,
	1, 2, 0, 3,   1, 3, 3, 3,   1, 2, 3, 3,   3, 3, 2, 2,
	1, 2, 0, 3,   2, 3, 3, 3,   0, 1, 3, 3,   3, 2, 1, 0,
	
	// Mode 4
	
	1, 2, 0, 3,   2, 3, 3, 3,   0, 1, 3, 3,   3, 2, 1, 0,
	1, 2, 0, 3,   2, 1, 3, 3,   2, 0, 3, 3,   1, 1, 0, 0,
	1, 2, 0, 3,   1, 2, 3, 3,   0, 2, 3, 3,   2, 1, 1, 0,
	1, 2, 0, 3,   1, 1, 3, 3,   0, 0, 3, 3,   0, 0, 0, 0,
	1, 2, 0, 3,   2, 3, 3, 3,   0, 1, 3, 3,   3, 2, 1, 0,
	1, 2, 0, 3,   2, 1, 3, 3,   2, 0, 3, 3,   1, 1, 0, 0,
	1, 2, 0, 3,   1, 2, 3, 3,   0, 2, 3, 3,   2, 1, 1, 0,
	1, 2, 0, 3,   1, 1, 3, 3,   0, 0, 3, 3,   0, 0, 0, 0,
	1, 2, 0, 3,   2, 3, 3, 3,   0, 1, 3, 3,   3, 2, 1, 0,
	1, 2, 0, 3,   2, 1, 3, 3,   2, 0, 3, 3,   1, 1, 0, 0,
	1, 2, 0, 3,   1, 2, 3, 3,   0, 2, 3, 3,   2, 1, 1, 0,
	1, 2, 0, 3,   1, 1, 3, 3,   0, 0, 3, 3,   0, 0, 0, 0,
	1, 2, 0, 3,   2, 3, 3, 3,   0, 1, 3, 3,   3, 2, 1, 0,
	1, 2, 0, 3,   2, 1, 3, 3,   2, 0, 3, 3,   1, 1, 0, 0,
	1, 2, 0, 3,   1, 2, 3, 3,   0, 2, 3, 3,   2, 1, 1, 0,
	1, 2, 0, 3,   1, 1, 3, 3,   0, 0, 3, 3,   0, 0, 0, 0,
	
	1, 2, 0, 3,   2, 2, 3, 3,   1, 1, 3, 3,   3, 3, 3, 3,
	1, 2, 0, 3,   2, 0, 3, 3,   1, 0, 3, 3,   3, 2, 2, 1,
	1, 2, 0, 3,   1, 3, 3, 3,   1, 2, 3, 3,   3, 3, 2, 2,
	1, 2, 0, 3,   2, 3, 3, 3,   0, 1, 3, 3,   3, 2, 1, 0,
	1, 2, 0, 3,   2, 2, 3, 3,   1, 1, 3, 3,   3, 3, 3, 3,
	1, 2, 0, 3,   2, 0, 3, 3,   1, 0, 3, 3,   3, 2, 2, 1,
	1, 2, 0, 3,   1, 3, 3, 3,   1, 2, 3, 3,   3, 3, 2, 2,
	1, 2, 0, 3,   2, 3, 3, 3,   0, 1, 3, 3,   3, 2, 1, 0,
	1, 2, 0, 3,   2, 2, 3, 3,   1, 1, 3, 3,   3, 3, 3, 3,
	1, 2, 0, 3,   2, 0, 3, 3,   1, 0, 3, 3,   3, 2, 2, 1,
	1, 2, 0, 3,   1, 3, 3, 3,   1, 2, 3, 3,   3, 3, 2, 2,
	1, 2, 0, 3,   2, 3, 3, 3,   0, 1, 3, 3,   3, 2, 1, 0,
	1, 2, 0, 3,   2, 2, 3, 3,   1, 1, 3, 3,   3, 3, 3, 3,
	1, 2, 0, 3,   2, 0, 3, 3,   1, 0, 3, 3,   3, 2, 2, 1,
	1, 2, 0, 3,   1, 3, 3, 3,   1, 2, 3, 3,   3, 3, 2, 2,
	1, 2, 0, 3,   2, 3, 3, 3,   0, 1, 3, 3,   3, 2, 1, 0,
	
	// Mode 5
	
	1, 2, 0, 3,   2, 3, 3, 3,   0, 1, 3, 3,   3, 2, 1, 0,
	1, 2, 0, 3,   2, 1, 3, 3,   2, 0, 3, 3,   1, 1, 0, 0,
	1, 2, 0, 3,   1, 2, 3, 3,   0, 2, 3, 3,   2, 1, 1, 0,
	1, 2, 0, 3,   1, 1, 3, 3,   0, 0, 3, 3,   0, 0, 0, 0,
	1, 2, 0, 3,   2, 3, 3, 3,   0, 1, 3, 3,   3, 2, 1, 0,
	1, 2, 0, 3,   2, 1, 3, 3,   2, 0, 3, 3,   1, 1, 0, 0,
	1, 2, 0, 3,   1, 2, 3, 3,   0, 2, 3, 3,   2, 1, 1, 0,
	1, 2, 0, 3,   1, 1, 3, 3,   0, 0, 3, 3,   0, 0, 0, 0,
	1, 2, 0, 3,   2, 3, 3, 3,   0, 1, 3, 3,   3, 2, 1, 0,
	1, 2, 0, 3,   2, 1, 3, 3,   2, 0, 3, 3,   1, 1, 0, 0,
	1, 2, 0, 3,   1, 2, 3, 3,   0, 2, 3, 3,   2, 1, 1, 0,
	1, 2, 0, 3,   1, 1, 3, 3,   0, 0, 3, 3,   0, 0, 0, 0,
	1, 2, 0, 3,   2, 3, 3, 3,   0, 1, 3, 3,   3, 2, 1, 0,
	1, 2, 0, 3,   2, 1, 3, 3,   2, 0, 3, 3,   1, 1, 0, 0,
	1, 2, 0, 3,   1, 2, 3, 3,   0, 2, 3, 3,   2, 1, 1, 0,
	1, 2, 0, 3,   1, 1, 3, 3,   0, 0, 3, 3,   0, 0, 0, 0,
	
	1, 2, 0, 3,   2, 2, 3, 3,   1, 1, 3, 3,   3, 3, 3, 3,
	1, 2, 0, 3,   2, 0, 3, 3,   1, 0, 3, 3,   3, 2, 2, 1,
	1, 2, 0, 3,   1, 3, 3, 3,   1, 2, 3, 3,   3, 3, 2, 2,
	1, 2, 0, 3,   2, 3, 3, 3,   0, 1, 3, 3,   3, 2, 1, 0,
	1, 2, 0, 3,   2, 2, 3, 3,   1, 1, 3, 3,   3, 3, 3, 3,
	1, 2, 0, 3,   2, 0, 3, 3,   1, 0, 3, 3,   3, 2, 2, 1,
	1, 2, 0, 3,   1, 3, 3, 3,   1, 2, 3, 3,   3, 3, 2, 2,
	1, 2, 0, 3,   2, 3, 3, 3,   0, 1, 3, 3,   3, 2, 1, 0,
	1, 2, 0, 3,   2, 2, 3, 3,   1, 1, 3, 3,   3, 3, 3, 3,
	1, 2, 0, 3,   2, 0, 3, 3,   1, 0, 3, 3,   3, 2, 2, 1,
	1, 2, 0, 3,   1, 3, 3, 3,   1, 2, 3, 3,   3, 3, 2, 2,
	1, 2, 0, 3,   2, 3, 3, 3,   0, 1, 3, 3,   3, 2, 1, 0,
	1, 2, 0, 3,   2, 2, 3, 3,   1, 1, 3, 3,   3, 3, 3, 3,
	1, 2, 0, 3,   2, 0, 3, 3,   1, 0, 3, 3,   3, 2, 2, 1,
	1, 2, 0, 3,   1, 3, 3, 3,   1, 2, 3, 3,   3, 3, 2, 2,
	1, 2, 0, 3,   2, 3, 3, 3,   0, 1, 3, 3,   3, 2, 1, 0,
	
	// Mode 6
	
	1, 2, 0, 3,   2, 3, 3, 3,   0, 3, 3, 3,   3, 2, 1, 0,
	1, 2, 0, 3,   0, 3, 3, 3,   0, 3, 3, 3,   0, 0, 0, 0,
	1, 2, 0, 3,   2, 3, 3, 3,   0, 3, 3, 3,   3, 2, 1, 0,
	1, 2, 0, 3,   0, 3, 3, 3,   0, 3, 3, 3,   0, 0, 0, 0,
	1, 2, 0, 3,   2, 3, 3, 3,   0, 3, 3, 3,   3, 2, 1, 0,
	1, 2, 0, 3,   0, 3, 3, 3,   0, 3, 3, 3,   0, 0, 0, 0,
	1, 2, 0, 3,   2, 3, 3, 3,   0, 3, 3, 3,   3, 2, 1, 0,
	1, 2, 0, 3,   0, 3, 3, 3,   0, 3, 3, 3,   0, 0, 0, 0,
	1, 2, 0, 3,   2, 3, 3, 3,   0, 3, 3, 3,   3, 2, 1, 0,
	1, 2, 0, 3,   0, 3, 3, 3,   0, 3, 3, 3,   0, 0, 0, 0,
	1, 2, 0, 3,   2, 3, 3, 3,   0, 3, 3, 3,   3, 2, 1, 0,
	1, 2, 0, 3,   0, 3, 3, 3,   0, 3, 3, 3,   0, 0, 0, 0,
	1, 2, 0, 3,   2, 3, 3, 3,   0, 3, 3, 3,   3, 2, 1, 0,
	1, 2, 0, 3,   0, 3, 3, 3,   0, 3, 3, 3,   0, 0, 0, 0,
	1, 2, 0, 3,   2, 3, 3, 3,   0, 3, 3, 3,   3, 2, 1, 0,
	1, 2, 0, 3,   0, 3, 3, 3,   0, 3, 3, 3,   0, 0, 0, 0,
	
	1, 2, 0, 3,   0, 3, 3, 3,   0, 3, 3, 3,   3, 3, 3, 3,
	1, 2, 0, 3,   2, 3, 3, 3,   0, 3, 3, 3,   3, 2, 1, 0,
	1, 2, 0, 3,   0, 3, 3, 3,   0, 3, 3, 3,   3, 3, 3, 3,
	1, 2, 0, 3,   2, 3, 3, 3,   0, 3, 3, 3,   3, 2, 1, 0,
	1, 2, 0, 3,   0, 3, 3, 3,   0, 3, 3, 3,   3, 3, 3, 3,
	1, 2, 0, 3,   2, 3, 3, 3,   0, 3, 3, 3,   3, 2, 1, 0,
	1, 2, 0, 3,   0, 3, 3, 3,   0, 3, 3, 3,   3, 3, 3, 3,
	1, 2, 0, 3,   2, 3, 3, 3,   0, 3, 3, 3,   3, 2, 1, 0,
	1, 2, 0, 3,   0, 3, 3, 3,   0, 3, 3, 3,   3, 3, 3, 3,
	1, 2, 0, 3,   2, 3, 3, 3,   0, 3, 3, 3,   3, 2, 1, 0,
	1, 2, 0, 3,   0, 3, 3, 3,   0, 3, 3, 3,   3, 3, 3, 3,
	1, 2, 0, 3,   2, 3, 3, 3,   0, 3, 3, 3,   3, 2, 1, 0,
	1, 2, 0, 3,   0, 3, 3, 3,   0, 3, 3, 3,   3, 3, 3, 3,
	1, 2, 0, 3,   2, 3, 3, 3,   0, 3, 3, 3,   3, 2, 1, 0,
	1, 2, 0, 3,   0, 3, 3, 3,   0, 3, 3, 3,   3, 3, 3, 3,
	1, 2, 0, 3,   2, 3, 3, 3,   0, 3, 3, 3,   3, 2, 1, 0,
	
	// Mode 7
	
	0, 0, 0, 0,   2, 3, 3, 3,   2, 1, 3, 3,   3, 2, 1, 0,
	0, 0, 0, 0,   2, 1, 3, 3,   2, 0, 3, 3,   1, 1, 0, 0,
	0, 0, 0, 0,   2, 3, 3, 3,   2, 3, 3, 3,   3, 2, 1, 0,
	0, 0, 0, 0,   1, 3, 3, 3,   1, 0, 3, 3,   0, 0, 0, 0,
	0, 0, 0, 0,   2, 3, 3, 3,   2, 1, 3, 3,   3, 2, 1, 0,
	0, 0, 0, 0,   2, 1, 3, 3,   2, 0, 3, 3,   1, 1, 0, 0,
	0, 0, 0, 0,   2, 3, 3, 3,   2, 3, 3, 3,   3, 2, 1, 0,
	0, 0, 0, 0,   1, 3, 3, 3,   1, 0, 3, 3,   0, 0, 0, 0,
	0, 0, 0, 0,   2, 3, 3, 3,   2, 1, 3, 3,   3, 2, 1, 0,
	0, 0, 0, 0,   2, 1, 3, 3,   2, 0, 3, 3,   1, 1, 0, 0,
	0, 0, 0, 0,   2, 3, 3, 3,   2, 3, 3, 3,   3, 2, 1, 0,
	0, 0, 0, 0,   1, 3, 3, 3,   1, 0, 3, 3,   0, 0, 0, 0,
	0, 0, 0, 0,   2, 3, 3, 3,   2, 1, 3, 3,   3, 2, 1, 0,
	0, 0, 0, 0,   2, 1, 3, 3,   2, 0, 3, 3,   1, 1, 0, 0,
	0, 0, 0, 0,   2, 3, 3, 3,   2, 3, 3, 3,   3, 2, 1, 0,
	0, 0, 0, 0,   1, 3, 3, 3,   1, 0, 3, 3,   0, 0, 0, 0,
	
	0, 0, 0, 0,   1, 2, 3, 3,   1, 0, 3, 3,   3, 3, 3, 3,
	0, 0, 0, 0,   2, 0, 3, 3,   2, 0, 3, 3,   3, 2, 2, 2,
	0, 0, 0, 0,   0, 3, 3, 3,   0, 2, 3, 3,   3, 3, 2, 2,
	0, 0, 0, 0,   2, 3, 3, 3,   2, 1, 3, 3,   3, 2, 1, 0,
	0, 0, 0, 0,   1, 2, 3, 3,   1, 0, 3, 3,   3, 3, 3, 3,
	0, 0, 0, 0,   2, 0, 3, 3,   2, 0, 3, 3,   3, 2, 2, 2,
	0, 0, 0, 0,   0, 3, 3, 3,   0, 2, 3, 3,   3, 3, 2, 2,
	0, 0, 0, 0,   2, 3, 3, 3,   2, 1, 3, 3,   3, 2, 1, 0,
	0, 0, 0, 0,   1, 2, 3, 3,   1, 0, 3, 3,   3, 3, 3, 3,
	0, 0, 0, 0,   2, 0, 3, 3,   2, 0, 3, 3,   3, 2, 2, 2,
	0, 0, 0, 0,   0, 3, 3, 3,   0, 2, 3, 3,   3, 3, 2, 2,
	0, 0, 0, 0,   2, 3, 3, 3,   2, 1, 3, 3,   3, 2, 1, 0,
	0, 0, 0, 0,   1, 2, 3, 3,   1, 0, 3, 3,   3, 3, 3, 3,
	0, 0, 0, 0,   2, 0, 3, 3,   2, 0, 3, 3,   3, 2, 2, 2,
	0, 0, 0, 0,   0, 3, 3, 3,   0, 2, 3, 3,   3, 3, 2, 2,
	0, 0, 0, 0,   2, 3, 3, 3,   2, 1, 3, 3,   3, 2, 1, 0,
	
	// Mode 1 with high BG3 prio
	
	1, 2, 0, 3,   2, 2, 3, 3,   1, 1, 0, 3,   3, 3, 2, 1,
	1, 2, 0, 3,   3, 2, 3, 3,   3, 1, 0, 3,   3, 3, 2, 1,
	1, 2, 0, 3,   2, 3, 3, 3,   1, 3, 0, 3,   3, 3, 2, 1,
	1, 2, 0, 3,   3, 3, 1, 3,   2, 2, 0, 3,   1, 1, 1, 1,
	1, 2, 0, 3,   1, 1, 2, 3,   0, 0, 2, 3,   2, 2, 1, 0,
	1, 2, 0, 3,   2, 1, 3, 3,   2, 0, 2, 3,   2, 2, 1, 0,
	1, 2, 0, 3,   1, 2, 3, 3,   0, 2, 2, 3,   2, 2, 1, 0,
	1, 2, 0, 3,   2, 2, 3, 3,   1, 1, 1, 3,   0, 0, 0, 0,
	1, 2, 0, 3,   2, 2, 3, 3,   1, 1, 0, 3,   3, 3, 2, 1,
	1, 2, 0, 3,   3, 2, 3, 3,   3, 1, 0, 3,   3, 3, 2, 1,
	1, 2, 0, 3,   2, 3, 3, 3,   1, 3, 0, 3,   3, 3, 2, 1,
	1, 2, 0, 3,   3, 3, 1, 3,   2, 2, 0, 3,   1, 1, 1, 1,
	1, 2, 0, 3,   1, 1, 2, 3,   0, 0, 2, 3,   2, 2, 1, 0,
	1, 2, 0, 3,   2, 1, 3, 3,   2, 0, 2, 3,   2, 2, 1, 0,
	1, 2, 0, 3,   1, 2, 3, 3,   0, 2, 2, 3,   2, 2, 1, 0,
	1, 2, 0, 3,   2, 2, 3, 3,   1, 1, 1, 3,   0, 0, 0, 0,
	
	0, 1, 2, 3,   2, 2, 2, 3,   1, 1, 0, 3,   3, 3, 3, 3,
	0, 1, 2, 3,   2, 1, 1, 3,   2, 1, 0, 3,   3, 3, 2, 2, // !!
	0, 1, 2, 3,   1, 2, 1, 3,   1, 2, 0, 3,   3, 3, 2, 2, // !!
	0, 1, 2, 3,   2, 2, 0, 3,   1, 1, 0, 3,   3, 3, 2, 1,
	0, 1, 2, 3,   1, 1, 3, 3,   0, 0, 2, 3,   3, 3, 3, 3,
	0, 1, 2, 3,   2, 0, 3, 3,   1, 0, 0, 3,   3, 3, 2, 1,
	0, 1, 2, 3,   0, 2, 3, 3,   0, 1, 0, 3,   3, 3, 2, 1,
	0, 1, 2, 3,   2, 2, 3, 3,   1, 1, 0, 3,   3, 3, 2, 1,
	0, 1, 2, 3,   2, 2, 2, 3,   1, 1, 0, 3,   3, 3, 3, 3,
	0, 1, 2, 3,   2, 1, 1, 3,   2, 1, 0, 3,   3, 3, 2, 2, // !!
	0, 1, 2, 3,   1, 2, 1, 3,   1, 2, 0, 3,   3, 3, 2, 2, // !!
	0, 1, 2, 3,   2, 2, 0, 3,   1, 1, 0, 3,   3, 3, 2, 1,
	0, 1, 2, 3,   1, 1, 3, 3,   0, 0, 2, 3,   3, 3, 3, 3,
	0, 1, 2, 3,   2, 0, 3, 3,   1, 0, 0, 3,   3, 3, 2, 1,
	0, 1, 2, 3,   0, 2, 3, 3,   0, 1, 0, 3,   3, 3, 2, 1,
	0, 1, 2, 3,   2, 2, 3, 3,   1, 1, 0, 3,   3, 3, 2, 1
};

#endif
